@@ -7,7 +7,6 @@
     :license: BSD, see LICENSE for details.
 """
 
-import os
 import pprint
 import re
 import sys
@@ -16,6 +15,7 @@ from path import path
 
 import docutils.nodes
 import sphinx
+from sphinx import addnodes
 from docutils.parsers.rst import Directive
 from docutils.transforms import Transform
 
@@ -41,7 +41,6 @@ SUPORTED_NODETYPES_POSITIONS = {
 
 existing_ids = set()
 inherits = {}
-doctree_pages = {}
 inherit_types = set()
 
 
@@ -66,52 +65,6 @@ def get_node_type(node):
     return str(type(node)).split('.')[-1].rstrip("'>")
 
 
-def generate_inheritref(node):
-    found = False
-    value = node.astext()
-    value = value.replace('[+id]', '')
-    value = value.strip()
-    node_type = get_node_type(node)
-    inherit_types.add(node_type)
-    source = node.document and node.document.attributes['source'] or ''
-    splitted = source.split(os.path.sep)
-    if len(splitted) >= 2:
-        prefix = '%s/%s' % (splitted[-2], splitted[-1].rstrip('.rst'))
-    else:
-        prefix = ''
-    word_count = 7
-    while True:
-        words = slugify(value)
-        words = words.replace(':', '_')
-        words = words.replace('.', '_')
-        words = words.replace('_', ' ')
-        words = words.split()
-        identifier = '_'.join(words[:word_count])
-        identifier = '%s:%s:%s' % (prefix, node_type, identifier)
-
-        if identifier not in existing_ids:
-            found = True
-            break
-        word_count += 1
-        if word_count > len(words):
-            break
-
-    if not found:
-        counter = 1
-        while True:
-            new_identifier = identifier + '_%d' % counter
-            if new_identifier not in existing_ids:
-                identifier = new_identifier
-                break
-            counter += 1
-
-    # TODO: By now we do not want the identifier to change because create_id()
-    # is called more than once. We need to find a better solution for creating
-    # the paragraph ID.
-    # existing_ids.add(identifier)
-    return identifier
-
-
 class InheritRef(Directive):
     has_content = True
     # required_arguments = 0
@@ -202,14 +155,16 @@ class Replacer(Transform):
                         % (ref, next_node.tagname))
 
     def apply(self):
-        config = self.document.settings.env.config
-        pattern = config.inheritance_pattern
+        env = self.document.settings.env
+        pattern = env.config.inheritance_pattern
         if isinstance(pattern, basestring):
             pattern = re.compile(pattern)
 
         def add_to_inherit_vals(inherit_vals, node):
             if not inherit_vals:
                 return
+            if isinstance(node, addnodes.toctree):
+                node['parent'] = inherit_vals['inherited_docname']
             if node.parent:
                 found = False
                 p = node.parent
@@ -233,7 +188,8 @@ class Replacer(Transform):
             if isinstance(node, (docutils.nodes.Inline, docutils.nodes.Text)):
                 continue
             if isinstance(node, inheritref_node):
-                inherit_node = Replacer._search_inherited_node(node, config)
+                inherit_node = Replacer._search_inherited_node(node,
+                    env.config)
                 inherit_node_type = node['inheritnodetype']
                 node['inheritnode'] = inherit_node
                 if (inherit_node_type not in ('section', 'title') or
@@ -270,7 +226,7 @@ class Replacer(Transform):
                 source = (node.document and node.document.attributes['source']
                     or '')
                 try:
-                    position, refsource, nodetype, refid = \
+                    position, inherited_docname, nodetype, refid = \
                             refdata.split(':')
                 except ValueError:
                     raise ValueError('Invalid inheritance ref "%s" at %s:%s'
@@ -284,7 +240,7 @@ class Replacer(Transform):
                             '%s:%s. The position "%s" is not supported for '
                             'node types "%s"' % (refdata, source, node.line,
                                 position, nodetype))
-                ref = '%s:%s:%s' % (refsource, nodetype, refid)
+                ref = '%s:%s:%s' % (inherited_docname, nodetype, refid)
                 current_inherit_ref = ref
 
                 container = inheritance_node()
@@ -293,6 +249,7 @@ class Replacer(Transform):
                 # container['rawsource'] = inherit_vals['rawsource']
 
                 current_inherit_vals = {
+                    'inherited_docname': inherited_docname,
                     'nodetype': nodetype,
                     'position': position,
                     'inheritance_node': container,
@@ -302,9 +259,17 @@ class Replacer(Transform):
                     'line': node.line,
                     }
                 inherits.setdefault(ref, []).append(current_inherit_vals)
-                if config.verbose:
-                    sys.stderr.write("Putting in inherits '%s' from file %s\n"
+                if env.config.verbose:
+                    sys.stderr.write("Put in inherits '%s' from file %s\n"
                             % (ref, source))
+
+                current_docname = path(source).relpath(env.srcdir)\
+                    .split('.')[0]
+                if env.config.verbose:
+                    sys.stderr.write("-- Adding to files_to_rebuild[%s]: %s\n"
+                        % (current_docname, inherited_docname))
+                env.files_to_rebuild.setdefault(current_docname,
+                    set()).add(inherited_docname)
                 continue
 
             add_to_inherit_vals(current_inherit_vals, node)
@@ -355,66 +320,127 @@ class Replacer(Transform):
 
 
 def init_transformer(app):
+    env = app.builder.env
+    srcdir_path = path(app.builder.env.srcdir)
+    modules = app.config.inheritance_modules
+    if isinstance(modules, (str, unicode)):
+        modules = [x.strip() for x in modules.split(',')]
+    app.config.exclude_patterns += [str(d.relpath(srcdir_path))
+        for d in srcdir_path.dirs() if d.relpath(srcdir_path) not in modules]
+
+    env.found_docs = set(d for d in env.found_docs
+        if d == app.config.master_doc or d.split('/')[0] in modules)
+
     if app.config.inheritance_plaintext:
         app.add_transform(Replacer)
 
 
-def check_module(app, docname, text):
-    modules = app.config.inheritance_modules
-    if isinstance(modules, (str, unicode)):
-        modules = [x.strip() for x in modules.split(',')]
-    path = os.path.split(docname)
-    if len(path) == 1:
+def sort_docnames(app, env, docnames):
+    inheritance_modules = app.config.inheritance_modules
+    if not inheritance_modules:
         return
-    module = path[-2]
-    if not module:
-        return
-    if module not in modules:
-        # If the module is not in the list of installed
-        # modules set as if the document was empty.
-        text[0] = ''
+
+    # inheritance modules should come ordered by inheritance dependency
+    # It reverse them to read before the document that inherits than the
+    # inherited document
+    inheritance_modules.reverse()
+
+    # docnames that aren't in any module. it should be only base index
+    original_docnames = docnames[:]
+    docnames = [d for d in docnames
+        if d.split('/')[0] not in inheritance_modules]
+    if len(docnames) > 1:
+        sys.stderr.write("WARN: there are more than one docname out of "
+            "inheritance modules: %s\n" % ",".join(docnames))
+
+    # add docnames sorted by order of its module in inheritance_modules list
+    docnames += [d
+        for (_, d_by_m) in zip(inheritance_modules,
+            [[d for d in original_docnames
+                    if d.split('/')[0] == m]
+                for m in inheritance_modules])
+        for d in d_by_m]
+
+
+def replace_inheritances(app, doctree):
+    if app.config.verbose:
+        sys.stderr.write("replace_inheritances: %s\n" % (doctree['source'], ))
+
+    search_inheritances(app, doctree)
+
+    if app.config.verbose:
+        sys.stderr.write("  end replace_inheritances(%s):\n"
+            % doctree['source'])
+
+    # Regenerate document toc to include inherit content
+    # The inheritance_modules list (in conf.py) must to be ordered by
+    # inheritance dependency to ensure that here all inheritances over current
+    # docname has been applied
+    if app.builder.env.docname in app.builder.env.toctree_includes:
+        # Avoid to duplicate content of toctree_includes[docname] when call
+        # build_toc_from
+        del app.builder.env.toctree_includes[app.builder.env.docname]
+    app.builder.env.build_toc_from(app.builder.env.docname, doctree)
 
 
 def search_inheritances(app, doctree):
     if not doctree:
         return
-    if not app.config.inheritance_autoreferences:
-        for node in doctree.traverse(inheritref_node):
-            inherit_node = node['inheritnode']
-            if node['inheritref'] not in inherits:
-                if app.config.inheritance_debug:
-                    sys.stderr.write("DEBUG: ref '%s' found in inheritref "
-                            "directive doesn't have any inheritance.\n"
-                                    % node['inheritref'])
-                continue
+    for node in doctree.traverse(inheritref_node):
+        inherit_node = node['inheritnode']
+        if node['inheritref'] not in inherits:
             if app.config.inheritance_debug:
-                sys.stderr.write("DEBUG: inheritance_%s (%s) found\n"
-                    % (node['inheritref'], node['inheritnodetype']))
-            apply_inheritance(app, [node, inherit_node], node['inheritref'])
-        return
-    # autoreferences
-    for node in doctree.traverse():
-        if isinstance(node, (docutils.nodes.Inline, docutils.nodes.Text)):
+                sys.stderr.write("DEBUG: ref '%s' found in inheritref "
+                        "directive doesn't have any inheritance.\n"
+                                % node['inheritref'])
             continue
-        # text = node.astext()
-        # source = node.document and node.document.attributes['source'] or ''
-        inheritref = generate_inheritref(node)
-        if inheritref in inherits:
-            apply_inheritance(app, [node], inheritref)
-    return
+        if app.config.inheritance_debug:
+            sys.stderr.write("DEBUG: inheritance_%s (%s) found\n"
+                % (node['inheritref'], node['inheritnodetype']))
+        apply_inheritance(app, [node, inherit_node], node['inheritref'])
 
 
 def apply_inheritance(app, node_list, inheritref):
     """
     ;param node_list: two-item list with the inheritref_node + inherited node
     """
+    def add_docname_to_files_to_rebuild(current_node_source,
+            inherited_node_source):
+        env = app.builder.env
+        node_docname = path(current_node_source).relpath(env.srcdir)\
+            .split('.')[0]
+        inherited_docname = path(inherited_node_source).relpath(env.srcdir)\
+            .split('.')[0]
+        if app.config.verbose:
+            sys.stderr.write("current/inherited_node_docname: %s--%s\n"
+                % (node_docname, inherited_docname))
+
+        # when document is included in toctree:
+        # - there is an entry for document with toctree in env.toctree_includes
+        #    with the list of documents included in toctree
+        # - there is the reverse relation in files_to_rebuild: per each file
+        #   included in toctree, there is an entry in files_to_rebuild with a
+        #   set of documents that include it
+        # Here, if inherited document is included in any toctree, it adds
+        # current document in env.files_to_rebuild for all documents that
+        # include inherited document
+        for docname in env.files_to_rebuild.get(inherited_docname, []):
+            if docname in env.toctree_includes:
+                env.files_to_rebuild.setdefault(docname,
+                    set()).add(node_docname)
+
     for n in node_list:
         if isinstance(n, (docutils.nodes.Inline, docutils.nodes.Text)):
             sys.stderr.write("WARNING: Inheritance to an unsuported Inline or "
                     "Text node. Ignored.\n")
             # if debug, more info
             return
-    for inherit_vals in inherits.get(inheritref, []):
+
+    inherit_vals_list = inherits.get(inheritref, [])
+    if not inherit_vals_list:
+        return
+    del inherits[inheritref]
+    for inherit_vals in inherit_vals_list:
         position = inherit_vals['position']
         inheritance_container = inherit_vals['inheritance_node']
 
@@ -423,9 +449,11 @@ def apply_inheritance(app, node_list, inheritref):
             sys.stderr.write("Searching inheritances in inherited nodes: %s\n"
                 % inheritance_container.__repr__())
         search_inheritances(app, inheritance_container)
+        if app.config.verbose:
+            sys.stderr.write("END searching inheritances in inherited nodes\n")
 
         if app.config.verbose:
-            sys.stderr.write("inherit_vals[%s]: %s" % (inheritref,
+            sys.stderr.write("inherit_vals[%s]: %s\n" % (inheritref,
                     pprint.pformat(inherit_vals)))
             sys.stderr.write("Applying inheritance %s over %s:\n"
                 " - parent of node_list[0] (%s): %s\n"
@@ -439,12 +467,20 @@ def apply_inheritance(app, node_list, inheritref):
             inherited_node = node_list[-1]
             inherited_node.parent.insert(
                 inherited_node.parent.index(inherited_node) + 1,
-                inheritance_container)
+                inheritance_container.children)
+            #    inheritance_container) it adds inheritance node which allows
+            # to generate specific HTML output but it breaks
+            # environment.build_toc_from() the 'traverse_in_section' part
+            add_docname_to_files_to_rebuild(inherit_vals['source'],
+                inherited_node.source)
         elif position == u'before':
             inherited_node = node_list[0]
             inherited_node.parent.insert(
                 inherited_node.parent.index(inherited_node),
-                inheritance_container)
+                inheritance_container.children)
+            # inheritance_container) IDEM
+            add_docname_to_files_to_rebuild(inherit_vals['source'],
+                inherited_node.source)
         elif position == u'inside':
             if inherit_vals['nodetype'] == 'toctree':
                 if app.config.verbose:
@@ -460,6 +496,7 @@ def apply_inheritance(app, node_list, inheritref):
                     inheritance_container[0]['entries'])
                 node_list[1]['includefiles'].extend(
                     inheritance_container[0]['includefiles'])
+                # TODO: add_docname_to_files_to_rebuild
             elif inherit_vals['nodetype'] == 'bullet_list':
                 if app.config.verbose:
                     sys.stderr.write("inheritance_container of 'inside' "
@@ -480,8 +517,13 @@ def apply_inheritance(app, node_list, inheritref):
                     for node in inheritance_container.children[1:]:
                         inherited_node.parent.insert(insert_index, node)
                         insert_index += 1
+                add_docname_to_files_to_rebuild(inherit_vals['source'],
+                    node_list[1].source)
             else:
-                node_list[-1].append(inheritance_container)
+                node_list[-1].extend(inheritance_container.children)
+                # node_list[-1].append(inheritance_container) Idem
+                add_docname_to_files_to_rebuild(inherit_vals['source'],
+                    node_list[-1].source)
 
         if app.config.verbose:
             sys.stderr.write("Applied inheritance %s over %s:\n"
@@ -497,53 +539,6 @@ def apply_inheritance(app, node_list, inheritref):
     return
 
 
-def replace_inheritances(app, doctree, fromdocname):
-    search_inheritances(app, doctree)
-    # regenerate the ToC of each page.
-    # TODO: it could be improved generating only for modified pages in
-    # inheritance
-    if app.config.verbose:
-        sys.stderr.write("Calling build_toc_from(fromdocname=%s)\n"
-                % fromdocname)
-    app.builder.env.build_toc_from(fromdocname, doctree)
-    return
-    # TODO: implement detection of modifications in inheritances to regenerate
-    # parents
-    modules = [m for m in app.config.inheritance_modules if m != 'trytond_doc']
-    if fromdocname == 'index' or fromdocname.split('/')[0] in modules:
-        if app.config.verbose:
-            sys.stderr.write("waiting to generate '%s' later\n" % fromdocname)
-        doctree_pages[fromdocname] = doctree
-
-
-def add_references(app, doctree, fromdocname):
-    if not app.config.inheritance_autoreferences:
-        return
-    for node in doctree.traverse():
-        if isinstance(node, (docutils.nodes.Inline, docutils.nodes.Text)):
-            # We do not want to consider inline nodes such as emphasis,
-            # strong or literal. Nor Text nodes which are the part of the
-            # paragraph that precede an inline node. There's already the
-            # Paragraph node and the anchor is added to it.
-            continue
-        if not node.parent:
-            continue
-        if isinstance(node, inheritref_node):
-            # to be safe. previously it must to be detected
-            if app.config.inheritance_debug:
-                sys.stderr.write("WARNING: Found 'inheritref' directive but "
-                        "'inheritance_autoreferences' option is True. "
-                        "Ignored.\n")
-            node.replace_self([])
-            continue
-
-        inheritref = inheritref_node('')
-        inheritref['inheritref'] = generate_inheritref(node)
-        inheritref['inheritnodetype'] = node.__class__.__name__
-        node.parent.insert(node.parent.index(node), inheritref)
-    return
-
-
 def report_warnings(app, exception):
     if app.config.verbose and inherit_types:
         sys.stderr.write("Used inherit types:\n - %s\n"
@@ -564,10 +559,12 @@ def report_warnings(app, exception):
 def visit_inheritref_node(self, node):
     inheritref = node['inheritref']
     target = docutils.nodes.target('', '', ids=[inheritref])
+    node.insert(0, target)
     self.visit_target(target)
     self.depart_target(target)
 
-    if self.settings.env.config.inheritance_debug:
+    if (hasattr(self, 'settings') and
+            self.settings.env.config.inheritance_debug):
         abbrnode = sphinx.addnodes.abbreviation('[+id]', '[+id]',
             explanation=inheritref)
         self.visit_abbreviation(abbrnode)
@@ -589,6 +586,14 @@ def depart_inheritance_node(self, node):
     self.body.append(self.context.pop())
 
 
+def visit_nothing(self, node):
+    pass
+
+
+def depart_nothing(self, node):
+    pass
+
+
 def setup(app):
     app.add_config_value('inheritance_plaintext', True, 'env')
     app.add_config_value('inheritance_pattern', re.compile(r'^#\:(.|[^#]+)#$'),
@@ -597,27 +602,25 @@ def setup(app):
             re.compile(r'(?P<prefix>[a-zA-Z/_]+):(?P<nodetype>[a-z_]+):'
                     '(?P<identifier>[a-zA-Z]+)'), 'env')
     app.add_config_value('inheritance_modules', [], 'env')
-    app.add_config_value('inheritance_autoreferences', False, 'env')
     app.add_config_value('inheritance_debug', False, 'env')
     app.add_config_value('verbose', False, 'env')
 
     app.add_node(inheritref_node,
         html=(visit_inheritref_node, depart_inheritref_node),
-        latex=(visit_inheritref_node, depart_inheritref_node),
-        text=(visit_inheritref_node, depart_inheritref_node),
-        man=(visit_inheritref_node, depart_inheritref_node),
-        texinfo=(visit_inheritref_node, depart_inheritref_node))
+        latex=(visit_nothing, depart_nothing),
+        text=(visit_nothing, depart_nothing),
+        man=(visit_nothing, depart_nothing),
+        texinfo=(visit_nothing, depart_nothing))
     app.add_node(inheritance_node,
         html=(visit_inheritance_node, depart_inheritance_node),
-        latex=(visit_inheritance_node, depart_inheritance_node),
-        text=(visit_inheritance_node, depart_inheritance_node),
-        man=(visit_inheritance_node, depart_inheritance_node),
-        texinfo=(visit_inheritance_node, depart_inheritance_node))
+        latex=(visit_nothing, depart_nothing),
+        text=(visit_nothing, depart_nothing),
+        man=(visit_nothing, depart_nothing),
+        texinfo=(visit_nothing, depart_nothing))
 
     app.add_directive('inheritref', InheritRef)
 
     app.connect(b'builder-inited', init_transformer)
-    app.connect(b'source-read', check_module)
-    app.connect(b'doctree-resolved', replace_inheritances)
-    app.connect(b'doctree-resolved', add_references)
+    app.connect(b'env-before-read-docs', sort_docnames)
+    app.connect(b'doctree-read', replace_inheritances)
     app.connect(b'build-finished', report_warnings)
